const config = {
    artboard: {
        minimumZoom: 0.01,
        maximumZoom: 200,
        incrementZoom: 0.01,
        maximumPixelDensity: 3.0,
        incrementPixelDensity: 0.01,
        friendlyNames: {
            'backgroundColor': 'Background<br>Color',
            'x': 'Position X',
            'y': 'Position Y',
            'width': 'Artboard Width',
            'height': 'Artboard Height',
            'zoom': 'Zoom',
            'pixelDensity': 'Preview<br>Resolution',
        },
    },
    layer: {
        defaultFonts: ['Version-2', 'TonkaVF'],
        letterDelayProps: ['fontSize_px', 'letterSpacing', 'color', 'fontVariationAxes'],
        autoCreateFirstLayer: true,
        defaultTexts: ['text', 'variable time', 'hello world'],
        panelOrder: [
            'fontFamily',
            'textAlignButtonsHorizontal',
            'textAlignment',
            'width',
            'height',
            'fontSize_px',
            'letterSpacing',
            'lineHeight',
            'text',
            'fontVariationAxes',
            'x',
            'y',
            'alignButtonsHorizontal',
            'alignButtonsVertical',
            'rotation',
            'transformOrigin',
            'mirror_x',
            'mirror_x_distance',
            'mirror_y',
            'mirror_y_distance',
            'mirror_xy',
            'color',
            'letterDelays',
        ],
        orderSpacing: 2,
        friendlyNames: {
            'fontFamily': 'Font Family',
            'textAlignButtonsHorizontal': '',
            'textAlignment': 'Text Alignment',
            'width': 'Wrapper Width',
            'height': '',
            'fontSize_px': 'Font Size',
            'letterSpacing': 'Letter Spacing',
            'lineHeight': 'Line Height',
            'text': 'Text',
            'fontVariationAxes': 'Variable Axes',
            'x': 'Position X',
            'y': 'Position Y',
            'alignButtonsHorizontal': '',
            'alignButtonsVertical': '',
            'rotation': 'Rotation',
            'transformOrigin': 'Rotation Origin',
            'mirror_x': 'Mirror X',
            'mirror_x_distance': 'X Distance',
            'mirror_y': 'Mirror Y',
            'mirror_y_distance': 'Y Distance',
            'mirror_xy': 'Mirrox XY',
            'color': 'Color',
            'letterDelays': 'Letter Delays',
        },
    },
    tp: {
        addKeyframesTimeout_s: 0.01,
    },
    projects: {
        savePrefix: 'vte_project_'
    },
    interactor: {
        zoomBaseFactor: 0.001,
        zoomDynamicMax: 42,
    },
    audio: {
        ignoreProps: ['transformOrigin', 'fontFamily', 'text', 'mirror_x', 'mirror_y', 'mirror_xy'],
        defaultSmoothing: 0.7,
        fftBandsAnalysed: 256 * 8,
        fftBandsUsed: 256 / 2,
        fftHeight: 256 / 4,
        colorSeparateRGBA: true,
        ignoreOutboundFrequencies: true,
        pitchCombineFrequencies: false,
    },
    record: {
        ignoreProps: ['transformOrigin', 'fontFamily', 'text', 'mirror_x', 'mirror_y', 'mirror_xy'],
        recordMapped: true,
    },
    midi: {
        touchTimeThreshold_s: 0.5,
        smoothingMix: 0.1,
    },
    fs: {
        idbfsDir: '/idbfs',
        idbfsFontDir: '/idbfs/fonts',
        idbfsTmpDir: '/idbfs/tmp',
    },
    timeline: {
        rolloverReset: true,
        rolloverThreshold_s: 0.02,
    },
    autoSave: true,
};

const Config = function() {
    const configKeys = Object.keys(config);
    for (let c = 0; c < configKeys.length; c++) {
        const key = configKeys[c];
        this[key] = config[key];
    }
};

export {
    Config
}
